#include "bucket.h"
#include "msmodel.h"
#include "scene.h"

Bucket::Bucket( Model *model, Vec3 position, dMatrix3 orientation, Vec3 scale, double mass, bool on, bool stationary, bool movable, const bool rotatable[3] ) :
    Object( model, position, orientation, scale, mass, on, stationary, movable, rotatable )
{
    Vec3 size = _model->get_max() - _model->get_min();

    size = vec_mult( _scale, size );

    dMassSetZero( &_mass );

    dGeomID geom_encap[5];  // encapsulated geometries
    dMass mass_encap;    // encapsulated mass

    // FIXME: this should be diff for each side if we do non-uniform
    // scaling
    Vec3f sides( size[0], 0.1f, size[2] );

    // centers of each side
    Vec3f pos[5];
    pos[0] = Vec3( 0.0, -size[1] / 2.0, 0.0 ); // bottom
    pos[1] = Vec3( -size[0] / 2.0, 0.0, 0.0 ); // left
    pos[2] = Vec3( size[0] / 2.0, 0.0, 0.0 ); // right
    pos[3] = Vec3( 0.0, 0.0, size[2] / 2.0 ); // front
    pos[4] = Vec3( 0.0, 0.0, -size[2] / 2.0 ); // back

    float rot_angle[5];
    Vec3f rot_axis[5];

    rot_angle[0] = 0.0f;
    rot_axis [0] = Vec3f( 0.0f, 0.0f, 0.0f );
    rot_angle[1] = DEG_TO_RAD( 90.0f );
    rot_axis [1] = Vec3f( 0.0f, 0.0f, 1.0f );
    rot_angle[2] = DEG_TO_RAD( -90.0f );
    rot_axis [2] = Vec3f( 0.0f, 0.0f, 1.0f );
    rot_angle[3] = DEG_TO_RAD( 90.0f );
    rot_axis [3] = Vec3f( 1.0f, 0.0f, 0.0f );
    rot_angle[4] = DEG_TO_RAD( -90.0f );
    rot_axis [4] = Vec3f( 1.0f, 0.0f, 0.0f );

    int i;
    for ( i = 0; i < 5; i++ ) {
        _geom_ids[i] = dCreateGeomTransform( get_space_id() );

        // destroy the encapsulated when this one is.
        dGeomTransformSetCleanup( _geom_ids[i], 1 );

        geom_encap[i] = dCreateBox( 0, sides[0], sides[1], sides[2] );
        dMassSetBoxTotal( &mass_encap, mass / 5.0, sides[0], sides[1], sides[2] );

        dGeomTransformSetGeom( _geom_ids[i], geom_encap[i] );
        dGeomTransformSetInfo( _geom_ids[i], 1 );

        // set the transformation (adjust the mass too)
        dGeomSetPosition( geom_encap[i], pos[i][0], pos[i][1], pos[i][2]);
        dMassTranslate( &mass_encap, pos[i][0], pos[i][1], pos[i][2] );
        dMatrix3 rot;
        dRFromAxisAndAngle( rot, rot_axis[i][0], rot_axis[i][1], rot_axis[i][2], rot_angle[i] );
        dGeomSetRotation( geom_encap[i], rot );

        dMassRotate( &mass_encap, rot );

        // add to the total mass
        dMassAdd( &_mass, &mass_encap );
    }

    // not sure if this is needed
//    dMassTranslate( &_mass, -_mass.c[0], -_mass.c[1], -_mass.c[2] );

    for ( i = 0; i < 5; i++) {
        if ( _geom_ids[i] ) {
            dGeomSetBody( _geom_ids[i], _body_id );
        }
    }

    dBodySetMass( _body_id, &_mass );
}

Bucket::~Bucket( )
{
    for ( int i = 0; i < 5; i++ ) {
        dGeomDestroy( _geom_ids[i] );
    }
}

void draw_geom( dGeomID g )
{
    if ( !g ) return;

    const dReal* pos = dGeomGetPosition (g);
    dReal rot[ 16 ];
    dMatrix3ToOpenGL( rot, dGeomGetRotation( g ) );

    glPushMatrix();
    glTranslatef( pos[0], pos[1], pos[2] );
    glMultMatrixf( rot );

    int type = dGeomGetClass (g);
    if ( type == dBoxClass ) {
        dVector3 sides;
        dGeomBoxGetLengths( g, sides );

        draw_solid_cube( -Vec3( sides ) / 2.0, Vec3( sides ) / 2.0, NULL );
    }
    else if ( type == dSphereClass ) {
        double radius = dGeomSphereGetRadius( g );

        GLUquadric *qobj = gluNewQuadric();
        gluQuadricTexture( qobj, GL_TRUE );
        gluSphere( qobj, radius, 20, 10 );
        gluDeleteQuadric( qobj );
    }
    else if ( type == dCCylinderClass ) {
        dReal radius, length;
        dGeomCCylinderGetParams( g, &radius, &length );

        draw_capped_cylinder( length, radius );
    }
    else if ( type == dGeomTransformClass ) {
        dGeomID g2 = dGeomTransformGetGeom (g);

        draw_geom( g2 );
    }

    glPopMatrix();
}

void Bucket::render( bool selected )
{
    Object::render( selected );

    for ( int i = 0; i < 5; i++ ) {
        draw_geom( _geom_ids[i] );
    }
}
