/**
 * @file object.cpp
 * Implementation of Object class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */

#include "object.h"
#include "model.h"
#include "scene.h"

#include <gfx/gl.h>

#include <sstream>

Object::Object( Model *model, Vec3 position, dMatrix3 orientation, Vec3 scale, double mass, bool on, bool stationary, bool movable, const bool rotatable[3]) :
    _model( model ),
    _scale( scale ),
    _geom_id( 0 ),
    _stationary( stationary ),
    _movable( movable ),
    _ethereal( false ),
    _on( on )
{
    // just so lots of other functions dont crash
    if( !model ) {
        model = new Model();
    }

    _body_id = dBodyCreate( get_world_id() );

    dBodySetPosition( _body_id, position[ 0 ], position[ 1 ], position[ 2 ] );
    _init_position = position;

    dBodySetData( _body_id, this );

    if( orientation == NULL ) {
        dQSetIdentity( _init_orientation );
    }
    else {
        dRtoQ( orientation, _init_orientation );
    }

    dBodySetQuaternion( _body_id, _init_orientation );

    _sim_on = _on;

    set_stationary( stationary );

    dBodySetData( _body_id, this );

    _rotatable[0] = rotatable[0];
    _rotatable[1] = rotatable[1];
    _rotatable[2] = rotatable[2];

    set_bbox_color( Vec3( 0.0, 1.0, 0.0 ) );
}

Object::~Object( void )
{
    // remove this object from ode
    if ( _body_id ) {
        dBodyDestroy( _body_id );
    }
    if ( _geom_id ) {
        dGeomDestroy( _geom_id );
    }

    // I assume the object will gain control of the model, and can delete
    // it at will.
    delete _model;
}

EffectHandler Object::get_effect_handler( Effect *effect )
{
//    if ( IS_PHYSICAL_EFFECT( effect ) && !is_ethereal() ) {
//        return (EffectHandler)&Object::_handle_physical_effect;
//    }
    return NULL; // no handler for this effect
}

void Object::_handle_physical_effect( PhysicalEffect *effect )
{
    /** @FIXME I don't know how to do rotations. Figure out where the
     * effect is acting, calculate rotation, etc
     */
    /// @FIXME collisions?

    /*
    if ( !is_stationary() ) {
        _net_force += effect->get_force();
    }
    */
}

void Object::render( bool selected )
{
    glPushMatrix();

    apply_transform();

    _model->render();

    glPopMatrix();
}

void Object::draw_decor( const Vec3& color )
{
    glPushMatrix();

    apply_transform();

    _model->draw_decor( color );

    glPopMatrix();
}

void Object::apply_transform( void )
{
    Vec3 pos = dBodyGetPosition( _body_id );

    glTranslated( pos[ 0 ], pos[ 1 ], pos[ 2 ] );

#if 1
    dReal rot[ 16 ];
    dMatrix3ToOpenGL( rot, dBodyGetRotation( _body_id ) );

    glMultMatrixf( rot );
#else
    // I'm pretty sure quaternions are faster
    // update: unfortunately, they break sometimes.
    const dReal *quat = dBodyGetQuaternion( _body_id );
    double rot_angle = RAD_TO_DEG( 2.0 * acos( quat[0] ) );

    glRotatef( rot_angle, quat[1], quat[2], quat[3] );
#endif

    glScaled( _scale[ 0 ], _scale[ 1 ], _scale[ 2 ] );
}

void Object::start_sim( void )
{
    dRealPtrToVec3( _init_position, dBodyGetPosition( _body_id ) );

    dRealPtrTodQuaternion( _init_orientation, dBodyGetQuaternion( _body_id ) );

    _sim_on = _on;
}

void Object::stop_sim( void )
{
    dBodySetPosition( _body_id, _init_position[ 0 ], _init_position[ 1 ], _init_position[ 2 ] );

    dBodySetQuaternion( _body_id, _init_orientation );

    dBodySetLinearVel( _body_id, 0.0, 0.0, 0.0 );
    dBodySetAngularVel( _body_id, 0.0, 0.0, 0.0 );

    _on = _sim_on;

    // are these necessary?
    dBodySetForce( _body_id, 0.0, 0.0, 0.0 );
    dBodySetTorque( _body_id, 0.0, 0.0, 0.0 );
}

void Object::setup_geom( void )
{
    dBodySetMass( _body_id, &_mass );

    dGeomSetBody( _geom_id, _body_id );

    // CHANGE: objects are now part of body's data (for multi-geom
    // bodies).
    // also, setting the geom's position is redundant (body takes care of
    // it).
#if 0
    dGeomSetPosition( _geom_id, _init_position[ 0 ], _init_position[ 1 ], _init_position[ 2 ] );

    dGeomSetQuaternion( _geom_id, _init_orientation );

    dGeomSetData( _geom_id, this );
#endif
}

void Object::collide( dContact* contact )
{
}

string Object::get_info( ) const
{
    ostringstream out;

    out << "type=\"" << get_type_str() << "\" ";
    out << get_model()->get_info() << " ";
    out << endl;
    out << "position=\"" << get_position() << "\" ";
    out << "forward=\"" << get_forward_vec() << "\" ";
    out << "up=\"" << get_up_vec() << "\" ";
    out << endl;
    out << "scale=\"" << get_scale() << "\" ";
    out << "mass=\"" << get_mass() << "\" ";
    out << endl;
    out << "on=\"" << to_string( is_on() ) << "\" ";
    out << "stationary=\"" << to_string( is_stationary() ) << "\" ";
    out << endl;
    out << "movable=\"" << to_string( is_movable() ) << "\" ";
    out << "rotatable=\""
        << ( is_rotatable( 0 ) ? 1 : 0 ) << " "
        << ( is_rotatable( 1 ) ? 1 : 0 ) << " "
        << ( is_rotatable( 2 ) ? 1 : 0 ) << "\" ";

    return out.str();
}
