/**
 * @file object.h
 * Header for Object class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _OBJECT_H
#define _OBJECT_H

#include "main.h"
#include "actor.h"
#include "odeutils.h"

#include <gfx/vec4.h>
#include <gfx/mat4.h>
#include <gfx/quat.h>

#include <ode/ode.h>

class Object;
class PhysicalEffect;
class Effect;
class Model;
class Camera;

/**
 * A method pointer for effect handlers (what gets called when the Object
 * recieves a certain effect).
 */
typedef void (Object::*EffectHandler)(Effect *);

/**
 * A physical, visible object.
 *
 * Objects are physical things in the scene, and interact with each other
 * via effects.  To receive effects, an Object must implement
 * get_effect_handler(), and return the function it wants called whenever
 * it receives a effect of a given type.
 */
class Object : public Actor {
public:
    /**
     * Constructor.
     *
     * @param model the model (graphical representation) of this object.
     * The object *will* obtain possession of the model (and *will*
     * delete it when destructed).
     * @param position 3d position of the object.
     * @param orientation 3d orientation of the object
     * @scale scaling factor of the object
     * @param mass the mass in kg.
     */
    Object( Model *model, Vec3 position = Vec3( 0.0, 0.0, 0.0 ), dMatrix3 orientation = NULL, Vec3 scale = Vec3( 1.0, 1.0, 1.0 ), double mass = 1.0, bool on = false, bool stationary = true, bool movable = false, const bool rotatable[3] = NULL);
    virtual ~Object( void );

    /**
     * Have the object determine what kind of Effect effect is, and return a
     * method pointer that will handle applications of that effect.
     *
     * @param effect the effect that we wish to find the handler for.
     * @return the handler (method pointer) for the given effect.
     */
    virtual EffectHandler get_effect_handler( Effect *effect );

    /**
     * Render the object to the screen.  I think this will actually just
     * call the model's render() function.
     */
    virtual void render( bool selected = false );

    /**
     * Draw a decoration around the object to show its status.
     */
    void draw_decor( const Vec3& color );

    virtual void apply_transform( void );

    virtual void tick( double t ) {}

    // keep the object's "sim" positions/orientations updated, and reset them
    // when necessary
    virtual void start_sim( void );
    virtual void stop_sim( void );

    /// notify the object that he has been collided with.
    virtual void collide( dContact* contact );

    //@{
    /**
     * Accessor functions.
     *
     * These give access to private object data.  See variables for
     * details.
     */
    void set_id( int id ) { _id = id; }
    void set_position( const Vec3& pos ) { dBodySetPosition( _body_id, pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void set_orientation_mat( const dMatrix3 orientation ) { dBodySetRotation( _body_id, orientation ); }
    void set_orientation_quat( const dQuaternion orientation ) { dBodySetQuaternion( _body_id, orientation ); }
    //void set_mass( double m ) { _mass = m; }
    void set_scale( const Vec3 s ) { _scale = s; }
    void set_stationary( const bool s ) { if( ( _stationary = s ) ) dBodyDisable( _body_id ); else dBodyEnable( _body_id ); }
    void set_movable( const bool s ) { _movable = s; }
    void set_rotatable( const bool s[3] )
        { _rotatable[0] = s[0]; _rotatable[1] = s[1]; _rotatable[2] = s[2]; }
    void set_ethereal( const bool e ) { _ethereal = e; }
    void set_on( const bool o ) { _on = o; }
    void set_bbox_color( const Vec3& v ) { _bbox_color = v; }

    int get_id( ) const { return _id; }
    Vec3 get_position( void ) const { const float* pos = dBodyGetPosition( _body_id ); return Vec3( pos[ 0 ], pos[ 1 ], pos[ 2 ] ); }
    void get_orientation_mat( dMatrix3 dmat3 ) const { dRealPtrTodMatrix3( dmat3, dBodyGetRotation( _body_id ) ); }
    Mat4 get_orientation_mat( ) const { Mat4 m; dMatrix3ToMat4( m, dBodyGetRotation( _body_id ) ); return m; }
    void get_orientation_quat( dQuaternion dquat ) const { dRealPtrTodQuaternion( dquat, dBodyGetQuaternion( _body_id ) ); }
    Quat get_orientation_quat( ) const { Quat q; dQuaternion_to_quat( q, dBodyGetQuaternion( _body_id ) ); return q; }
    Vec3 get_forward_vec( ) const { return get_orientation_mat() * Vec3( 0, 0, 1 ); }
    Vec3 get_up_vec( ) const { return get_orientation_mat() * Vec3( 0, 1, 0 ); }
    virtual double get_mass( void ) const { dMass dm; dBodyGetMass( _body_id, &dm ); return dm.mass; }
    virtual Vec3 get_scale( void ) const { return _scale; }
    virtual bool is_stationary( void ) const { return _stationary; }
    virtual bool is_movable( void ) const { return _movable; }
    virtual bool is_rotatable( int i ) const { return _rotatable[i]; }
    virtual bool is_ethereal( void ) const { return _ethereal; }
    virtual bool is_on( void ) const { return _on; }
    virtual Vec3 get_bbox_color( ) const { return _bbox_color; }
    virtual Model *get_model( ) const { return _model; }
    virtual string get_type_str( ) const { return "object"; }
    virtual string get_info( ) const;

    Vec3 get_init_position( void ) const { return _init_position; }
    void get_init_orientation_mat( dMatrix3 dmat3 ) const { dQtoR( _init_orientation, dmat3 ); }
    void get_init_orientation_quat( dQuaternion dquat ) const { dRealPtrTodQuaternion( dquat, _init_orientation ); }
    bool is_sim_on( void ) const { return _sim_on; }

    const dBodyID get_body_id() { return _body_id; }
    const dGeomID get_geom_id() { return _geom_id; }

    void enable_body( void ) { dBodyEnable( _body_id ); }
    void disable_body( void ) { dBodyDisable( _body_id ); }

    //@}
protected:
    Model* _model;              ///< my graphical representation

    Vec3 _scale;                ///< the scaling factor

    int _id;

    dBodyID _body_id;
    dGeomID _geom_id;

    Vec3 _init_position;
    dQuaternion _init_orientation;
    bool _sim_on;

    dMass _mass;

    void setup_geom( void );

    bool _stationary;           ///< does the physical simulation affect me?
    bool _movable;              ///< can the user move me?
    bool _rotatable[3];         ///< can the user rotate me about x,y,z?
    bool _ethereal;             ///< can objects collide with me / can I overlap objects

    bool _on;                   ///< am i activated?

    Vec3 _bbox_color;
private:

    /**
     * Default handler for physical effects (ie, forces).
     *
     * Most objects will not have a specialized handler for Physical
     * effects, in which case those effects will be handled here, via
     * Newton's laws.
     * 
     * @param effect the effect that we're receiving.
     */
    void _handle_physical_effect( PhysicalEffect *effect );

};

#endif // header guard
