/**
 * @file actors/shapeobject.cpp
 * Implementation of Seesaw class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "seesaw.h"
#include "scene.h"
#include "model.h"
#include "utils.h"

#include <gfx/gl.h>

Seesaw::Seesaw( Model *model, Vec3 position, dMatrix3 orientation, Vec3 scale, double mass, bool on, bool stationary, bool movable, const bool rotatable[3] ) :
    ShapeObject( model, SHAPE_BOX, position, orientation, scale, mass, on, stationary, movable, rotatable )
{
    _hinge = dJointCreateHinge( get_world_id(), 0 );
    dJointAttach( _hinge, _body_id, 0 );
    dJointSetHingeAnchor( _hinge, _init_position[0], _init_position[1], _init_position[2] );
    dJointSetHingeAxis( _hinge, 1.0f, 0.0f, 0.0f );
    dJointSetHingeParam( _hinge, dParamLoStop, DEG_TO_RAD( -30.0f ) );
    dJointSetHingeParam( _hinge, dParamHiStop, DEG_TO_RAD( 30.0f ) );
}

Seesaw::~Seesaw()
{
    dJointDestroy( _hinge );
}

void Seesaw::render( bool selected )
{
    Object::render( selected );

    Vec3 size = _model->get_max() - _model->get_min();
    size[0] = size[0] / 2.0;
    size[1] = size[1] * 3.0;
    size[2] = size[2] / 10.0;

    Vec3 left_top( -size[0], 0.0, 0.0 );
    Vec3 left_bl(  -size[0], -size[1], size[2] );
    Vec3 left_br(  -size[0], -size[1], -size[2] );
    Vec3 right_top( size[0], 0.0, 0.0 );
    Vec3 right_bl(  size[0], -size[1], size[2] );
    Vec3 right_br(  size[0], -size[1], -size[2] );

    glPushMatrix();
    apply_transform();

    // draw the base
    glBegin( GL_QUADS );
    glTexCoord2d( 0.0, 1.0 );
    glVertex3dv( left_top );
    glTexCoord2d( 1.0, 1.0 );
    glVertex3dv( right_top );
    glTexCoord2d( 1.0, 0.0 );
    glVertex3dv( right_bl );
    glTexCoord2d( 0.0, 0.0 );
    glVertex3dv( left_bl );
    glEnd();

    glBegin( GL_QUADS );
    glTexCoord2d( 0.0, 1.0 );
    glVertex3dv( left_top );
    glTexCoord2d( 1.0, 1.0 );
    glVertex3dv( right_top );
    glTexCoord2d( 1.0, 0.0 );
    glVertex3dv( right_br );
    glTexCoord2d( 0.0, 0.0 );
    glVertex3dv( left_br );
    glEnd();

    glBegin( GL_TRIANGLES );
    glTexCoord2d( 0.0, 1.0 );
    glVertex3dv( left_top );
    glTexCoord2d( 0.0, 0.0 );
    glVertex3dv( left_br );
    glTexCoord2d( 1.0, 0.0 );
    glVertex3dv( left_bl );

    glTexCoord2d( 0.0, 1.0 );
    glVertex3dv( right_top );
    glTexCoord2d( 0.0, 0.0 );
    glVertex3dv( right_br );
    glTexCoord2d( 1.0, 0.0 );
    glVertex3dv( right_bl );
    glEnd();

    glPopMatrix();
}

void Seesaw::start_sim( void )
{
    Object::start_sim();

    dJointSetHingeAnchor( _hinge, _init_position[0], _init_position[1], _init_position[2] );

    Mat4 mat;
    dMatrix3ToMat4( mat, dBodyGetRotation( _body_id ) );
    Vec3 axis = vec_dropw( mat * Vec4( 1.0, 0.0, 0.0, 0.0 ) );

    dJointSetHingeAxis( _hinge, axis[0], axis[1], axis[2] );
}
