/**
 * @file actors/worldbox.cpp
 * Implementation of Actors/worldbox class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "worldbox.h"
#include "scene.h"
#include "model.h"

#include <sstream>

WorldBox::WorldBox( Vec3 size, const Material& box_mat, const Material& cloud_mat ) :
    _material( box_mat ),
    _clouds( cloud_mat ),
    _size( size )
{
    // create the box
    _min = -_size / 2.0;
    _max = _size / 2.0;

    _material.reload_texture();
    _clouds.reload_texture();

    _cloud_ds = 0.0;
    _cloud_dt = 0.0;

    // attach the bounding planes to the static environment (i.e. they dont
    // need rigid body id's)
    _geom_id[ 0 ] = dCreatePlane( get_space_id(),  1,  0,  0,  _min[0] );
    _geom_id[ 1 ] = dCreatePlane( get_space_id(), -1,  0,  0, -_max[0] );
    _geom_id[ 2 ] = dCreatePlane( get_space_id(),  0,  1,  0,  _min[1] );
    _geom_id[ 3 ] = dCreatePlane( get_space_id(),  0, -1,  0, -_max[1] );
    _geom_id[ 4 ] = dCreatePlane( get_space_id(),  0,  0,  1,  _min[2] );
    _geom_id[ 5 ] = dCreatePlane( get_space_id(),  0,  0, -1, -_max[2] );
}

WorldBox::~WorldBox( void )
{
    for ( int i = 0; i < 6; i++ ) {
        dGeomDestroy( _geom_id[i] );
    }
}

void WorldBox::render( void )
{
    // worldbox

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();

    _material.make_current();

    // figure out which sides of the box the camera is on, and don't draw
    // those sides
    bool draw_faces[6] = { true, true, true, true, true, true };
    Vec3 camera = SceneManager::get_scene()->get_camera()->get_position();
	
    if ( camera[0] < _min[0] )
        draw_faces[CUBE_LEFT] = false;
    else if ( camera[0] > _max[0] )
        draw_faces[CUBE_RIGHT] = false;
    if ( camera[1] < _min[1] )
        draw_faces[CUBE_BOTTOM] = false;
    else if ( camera[1] > _max[1] )
        draw_faces[CUBE_TOP] = false;
    if ( camera[2] < _min[2] )
        draw_faces[CUBE_FRONT] = false;
    else if ( camera[2] > _max[2] )
        draw_faces[CUBE_BACK] = false;	

    draw_solid_cube( _min, _max, draw_faces );

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();


    // clouds

    glMatrixMode( GL_TEXTURE );
    glPushMatrix();

    _clouds.make_current();
    _cloud_ds += 0.0003;
    _cloud_dt += 0.0001;
    glTranslated( _cloud_ds, _cloud_dt, 0.0 );
    glScaled( 10.0, 10.0, 10.0 );

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();

    glRotated( 45.0, 0.0, 0.0, 1.0 );
    glScaled( 1.0, 2.0, 1.0 );
    glRotated( -45.0, 0.0, 0.0, 1.0 );

    glBegin( GL_QUADS );
    const double size = 200.0;
    const double dist = 200.0;

    glNormal3d( 0, 0, -1 );
    glTexCoord2f( 0, 1 );
    glVertex3d( -size, -size, -dist );
    glTexCoord2f( 0, 0 );
    glVertex3d( -size, size, -dist );
    glTexCoord2f( 1, 0 );
    glVertex3d( size, size, -dist );
    glTexCoord2f( 1, 1 );
    glVertex3d( size, -size, -dist );

    glEnd();

    glMatrixMode( GL_TEXTURE );
    glPopMatrix();

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
}

string WorldBox::get_info( ) const
{
    ostringstream out;

    out << "size=\"" << _size << "\" ";
    out << "box=\"" << _material.texture_filename << "\" ";
    out << "clouds=\"" << _clouds.texture_filename << "\" ";

    return out.str();
}
