/**
 * @file effect.h
 * Header for Effect class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _EFFECTS_H
#define _EFFECTS_H

#include "main.h"
#include "object.h"

#include <map>
#include <list>
using namespace std;

/**
 * A message from one object to another that some effect is being exerted.
 *
 * All Effects will derive from this.
 */
class Effect {
public:

    /**
     * A list of "who cares about this effect".
     *
     * Each Effect will probably maintain a list of which Objects actually
     * respond to that type of effect, as an optimization (so we don't
     * waste time trying to apply CheeseEffect's to BowlingBalls).
     */
    typedef map<Object*, EffectHandler> RecList;

    /**
     * Dummy constructor, pretty much here solely for adding to the
     * receiver list.
     */
    Effect( void ) : _sender( NULL ) { }

    /**
     * Constructor for Effect.
     * @param sender who sent this effect.
     */
    Effect( Actor *sender ) : _sender( sender ) { }

    /**
     * Calculations needed prior to application of the effect.
     * 
     * Sometimes an Effect needs to do some calculations before it applies
     * itself to the receiver.  This function is called prior to each
     * application, allowing fine-tuning.
     * @see Gravity
     * @param who who we are about to apply the effect to
     */
    virtual void pre_apply( Object *who );

    /**
     * Apply this effect to the receiver, using the appropriate
     * EffectHandler.
     * @param who who to apply the effect to
     */
    virtual void apply_to( Object *who );

    /**
     * Apply this effect to everyone who cares about it, ie everyone in its
     * WhoCares list.
     */
    virtual void apply_to_all( void );

    /**
     * @return the WhoCares list.
     */
    virtual RecList& get_receiver_list( void ) = 0;

    /**
     * Add who's effect handler for this effect to our receiver list.
     * @NOTE this can't be a static function, because we need an actual
     * instance of some Effect class to pass to who's get_handler()
     * function.
     */
    virtual void add_handler(Object *who);

    /**
     * Clear all handlers.
     */
    virtual void clear_handlers();

    /**
     * @param who the receiver we're interesting in finding a handler for.
     * @return the EffectHandler that receiver who has designated as the
     * handler for this type of effect.
     */
    EffectHandler Effect::find_handler( Object *who )
        { return ( get_receiver_list() )[ who ]; }

    /**
     * @return who sent this effect.
     */
    Actor *get_sender( void ) { return _sender; }

    const Vec3 get_origin( void );

protected:

    Actor *_sender;     ///< who sent this effet.
};

#define TO_EFFECT(pobj, type) (dynamic_cast<type*>(pobj))

/**
 * A somewhat hackish way of automatically making a list of Effects.
 *
 * Each effect type in the hierarchy should create a dummy global instance
 * of this class, passing a pointer to a dummy global instance of the
 * Effect class.  The RegisterEffect constructor adds each pointer to a
 * list, which is then iterated through when we need to construct static
 * lists for each Effect in the hierarchy.
 * @see Scene::generate_handlers()
 */
class RegisterEffect {
public:
    typedef list<Effect*> EffectList;
    typedef EffectList::iterator iterator;

    /**
     * This adds the passed effect to the list of effects.
     *
     * Note: the effect_list must be NULL before this constructor is
     * called.
     * @see effect.cpp
     * @param effect the effect to add.
     */
    RegisterEffect( Effect *effect );

    /**
     * @return an iterator to the beginning of the list of effects.
     */
    static iterator begin( void );

    /**
     * @return an iterator to the end of the list of effects.
     */
    static iterator end( void );
};

#endif // header guard
