/**
 * @file gamestate/level.cpp
 * Implementation of Gamestate/level class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "level.h"
#include "model.h"
#include "msmodel.h"
#include "shapemodel.h"
#include "bucket.h"
#include "conveyorbelt.h"
#include "trigger.h"
#include "seesaw.h"

#include <fstream>

Shape get_shape( const CL_DomElement& element, string name )
{
    string shape_str = get_string( element, name );

    if ( shape_str == "box" ) {
        return SHAPE_BOX;
    }
    else if ( shape_str == "sphere" ) {
        return SHAPE_SPHERE;
    }
    else if ( shape_str == "cylinder" ) {
        return SHAPE_CYLINDER;
    }

    return SHAPE_BOX;
}

Object *Level::new_object_from_dom( const CL_DomElement& element )
{
    Object *obj = NULL;

    // all objects need at least these attributes
    if ( !element.has_attribute( "type" )
            || !element.has_attribute( "position" )
            || !element.has_attribute( "forward" )
            || !element.has_attribute( "up" ) ) {
        return NULL;
    }

    string type = get_string( element, "type" );
    Vec3 position = get_vec3( element, "position" );
    Vec3 forward = get_vec3( element, "forward" );
    Vec3 up = get_vec3( element, "up" );
    string model_filename = get_string( element, "model" );
    string texture_filename = get_string( element, "texture" );
    Shape shape = get_shape( element, "shape" );

    // put these only in the objects that need them.
    double mass = get_double( element, "mass" );
    Vec3 scale = get_vec3( element, "scale" );
    bool on = get_bool( element, "on" );
    bool stationary = get_bool( element, "stationary" );
    bool movable = get_bool( element, "movable" );
    int connection_id = get_int( element, "connectionID" );

    Vec3 rotatable_vec = get_vec3( element, "rotatable" );
    bool rotatable[3];
    rotatable[0] = rotatable_vec[0];
    rotatable[1] = rotatable_vec[1];
    rotatable[2] = rotatable_vec[2];

    Model *model;
    if ( !model_filename.empty() ) {
        model = new MSModel( "resources/models/", model_filename );
    }
    else if ( element.has_attribute( "shape" ) ) {
        model = new ShapeModel( shape, Vec3( 1.0 ) );
    }
    else {
        model = new Model;
    }

    if ( !texture_filename.empty() ) {
        Material mat = grey.copy_with_texture( texture_filename );
        model->set_material( mat );
    }

    dMatrix3 orientation;
//    Vec3 right = cross( forward, up );
    // oops.. they don't want a right vector, they want an x axis.
    // here I assume that an object's personal x,y,z axes are as follows:
    // forward = z, up = y, left = x
    Vec3 x_axis = up ^ forward;

    dRFrom2Axes( orientation, x_axis[ 0 ], x_axis[ 1 ], x_axis[ 2 ], up[ 0 ], up[ 1 ], up[ 2 ] );

    // TODO: I think we should pass in an ODE geom, or some kind of enum
    // that we can vary, so that our objects can have different geometries
    // without being different objects. Done.
    if( CL_String::compare_nocase( type, "bucket" ) ) {
        obj = new Bucket( model, position, orientation, scale, mass, on, stationary, movable, rotatable );
    }
    else if( CL_String::compare_nocase( type, "trigger" ) ) {
        Trigger *trig = new Trigger( model, shape, position, orientation, scale, mass, on, stationary, movable, rotatable );
        trig->set_connection( _objects[ connection_id ] );
        obj = trig;
    }
    else if( CL_String::compare_nocase( type, "shape" ) ) {
        obj = new ShapeObject( model, shape, position, orientation, scale, mass, on, stationary, movable, rotatable );
    }
    else if( CL_String::compare_nocase( type, "seesaw" ) ) {
        obj = new Seesaw( model, position, orientation, scale, mass, on, stationary, movable, rotatable );
    }
    else if( CL_String::compare_nocase( type, "conveyorbelt" ) ) {
        obj = new ConveyorBelt( model, shape, position, orientation, scale, mass, on, stationary, movable, rotatable );
    }
    else {
        cerr << "Object type not recognized" << endl;
        return NULL;
    }

    return obj;
}

Level::Level( string filename ) :
    _filename( filename )
{
    cout << "Loading: " << filename << endl;

    CL_ResourceManager level_resources( filename );

    _description = CL_String::load( "Description/LevelDescription", &level_resources );

    _id = CL_Integer( "Description/ID", &level_resources );

    CL_DomElement element;
    
    CL_Resource camera_resource = level_resources.get_resource( "World/camera" );
    camera_resource.load();
    element = camera_resource.get_element();
    _camera_radius = get_double( element, "radius" );
    _camera_elevation = get_double( element, "elevation" );
    _camera_azimuth = get_double( element, "azimuth" );

    CL_Resource world_resource = level_resources.get_resource( "World/world" );
    world_resource.load();
    element = world_resource.get_element();

    Vec3 world_size = get_vec3( element, "size" );
    string box_texture = get_string( element, "box" );
    string cloud_texture = get_string( element, "clouds" );

    Material box_mat = grey.copy_with_texture( box_texture );
    Material cloud_mat = grey.copy_with_texture( cloud_texture );

    _world_box = new WorldBox( world_size, box_mat, cloud_mat );

    list<string> goal_list = level_resources.get_resources_of_type( "goal", "Goals" );
    list<string>::iterator itr;

    for( itr = goal_list.begin(); itr != goal_list.end(); ++itr ) {
        CL_Resource goal_resource = level_resources.get_resource( *itr );

        goal_resource.load();

        CL_DomElement element = goal_resource.get_element();

        int ID = get_int( element, "ID" );

        string type = get_string( element, "type" );

        int object_id = get_int( element, "objectID" );

        Vec3 position( get_vec3( element, "position" ) );

        double radius = get_double( element, "radius" );

        double delay = get_double( element, "delay" );

        GoalType goal_type;
        if( CL_String::compare_nocase( type, "inside" ) )
            goal_type = INSIDE;
        else if( CL_String::compare_nocase( type, "outside" ) )
            goal_type = OUTSIDE;
        else if( CL_String::compare_nocase( type, "activate" ) )
            goal_type = ACTIVATE;
        else if( CL_String::compare_nocase( type, "deactivate" ) )
            goal_type = DEACTIVATE;
        else
            cerr << "Goal not recognized" << endl;

        cerr << "adding goal " << ID << " with obj id " << object_id << endl;
        _goals[ ID ] = new Goal( ID, goal_type, object_id, position, radius, delay );

        goal_resource.unload();
    }

    list<string> actor_list = level_resources.get_resources_of_type( "actor", "Actors" );

    for( itr = actor_list.begin(); itr != actor_list.end(); ++itr ) {
        CL_Resource actor_resource = level_resources.get_resource( *itr );

        actor_resource.load();

        CL_DomElement element = actor_resource.get_element();

        /*
        int ID = CL_String::to_int( element.get_attribute( "ID" ) );

        string type = element.get_attribute( "type" );

        if( CL_String::compare_nocase( type, "gravity" ) ) {
            actors[ ID ] = new Gravity( get_vec3( element, "force" ) );
        }
        else
            cerr << "Actor not recognized" << endl;
        */

        actor_resource.unload();
    }

    list<string> object_list = level_resources.get_resources_of_type( "object", "Objects" );

    for( itr = object_list.begin(); itr != object_list.end(); ++itr ) {
        CL_Resource object_resource = level_resources.get_resource( *itr );

        object_resource.load();

        CL_DomElement element = object_resource.get_element();

        int ID = CL_String::to_int( element.get_attribute( "ID" ) );

        Object *obj = new_object_from_dom( element );

        if ( obj ) {
            _objects[ ID ] = obj;
            obj->set_id( ID );
        }
        else {
            cerr << "Error parsing object element" << endl;
        }

        object_resource.unload();
    }
}

Level::~Level( void )
{
    delete _world_box;

    // make sure the objects are removed from ode before either
    // reloading them or loading new ones
    ObjectIterator itr;
    for( itr = _objects.begin(); itr != _objects.end(); ++itr ) {
        delete itr->second;
    }

    GoalIterator git;
    for( git = _goals.begin(); git != _goals.end(); ++git ) {
        delete git->second;
    }
}

void Level::tick( double t )
{
    GoalIterator itr;

    for( itr = _goals.begin(); itr != _goals.end(); ++itr ) {
        Goal* goal = itr->second;
        goal->tick( t );
    }
}

void Level::render( void )
{
    _world_box->render();
}

const bool Level::level_completed( void )
{
    bool completed = true;

    GoalIterator itr;

    for( itr = _goals.begin(); itr != _goals.end(); ++itr ) {
        Goal* goal = itr->second;

        // goals only have an object_id, not any more info about objects, so
        // we need to pass in the pointer to the object mentioned in the goal
        if( !goal->check_goal( _objects[ goal->get_object_id() ] ) ) {
            completed = false;
            break;
        }
    }


    return completed;
}

void Level::stop_sim( void )
{
    ObjectIterator itr;

    // reset the object's position and orientation to the position and
    // orientation it had before the simulation started
    for( itr = _objects.begin(); itr != _objects.end(); ++itr ) {
        itr->second->stop_sim();
    }
}

void Level::reset_level( void )
{
    // do any other initialization needed
}

void Level::save( string filename )
{
    if ( filename.empty() ) {
        save( cout );
    }
    else {
        ofstream fout( filename.c_str(), ios::out );
        save( fout );
        fout.close();
    }
}

void Level::save( ostream& out )
{
#define printline( x ) out << spaces << x << endl
#define indent() spaces += "  ";
#define unindent() spaces.erase( 0, 2 );

    string spaces = "";

    printline( "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>" );
    printline( "<resources>" );
    indent();

    
    printline( "<section name=\"Description\">" );
    indent();

    printline( "<integer name=\"ID\" value=\"" << _id << "\" />" );
    printline( "<string name=\"LevelDescription\" value=\"" << _description << "\" />" );

    unindent();
    printline( "</section>" );
    out << endl;


    printline( "<section name=\"World\">" );
    indent();

    printline( "<world name=\"world\" " << _world_box->get_info() << " />" );
    printline( "<camera name=\"camera\" "
            "radius=\"" << _camera_radius << "\" "
            "elevation=\"" << _camera_elevation << "\" "
            "azimuth=\"" << _camera_azimuth << "\" "
            "/>" );

    unindent();
    printline( "</section>" );
    out << endl;

    printline( "<section name=\"Goals\">" );
    indent();

    for( GoalIterator goal_itr = _goals.begin(); goal_itr != _goals.end(); ++goal_itr ) {
        int goal_id = (*goal_itr).first;
        Goal *goal = (*goal_itr).second;

        printline(
                "<goal "
                "name=\"" << goal_id << "\" "
                "ID=\"" << goal_id << "\" "
                << goal->get_info() <<
                "/>" );
    }

    unindent();
    printline( "</section>" );
    out << endl;

    printline( "<section name=\"Objects\">" );
    indent();

    for( ObjectIterator object_itr = _objects.begin(); object_itr != _objects.end(); ++object_itr ) {
        int obj_id = (*object_itr).first;
        Object* obj = (*object_itr).second;

        printline(
                "<object "
                "name=\"" << obj->get_type_str() << obj_id << "\" "
                "ID=\"" << obj_id << "\" "
                << endl
                << obj->get_info() <<
                "/>" );
    }

    
    unindent();
    printline( "</section>" );

    unindent();
    printline( "</resources>" );
}
