/**
 * @file scene/camera.cpp
 * Implementation of Scene/camera class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "camera.h"

#include <gfx/gl.h>

#include <math.h>

Camera::Camera(ProjectType pt) :
    _radius(10.0),
    _elevation(0.0),
    _azimuth(0.0),
    _project_type(pt)
{
    _min_radius = 5.0;
    _max_radius = 25.0;

    _min_elevation = 10.0;
    _max_elevation = 80.0;

    _up_vec = Vec3( 0.0, 1.0, 0.0 );
}

void Camera::set_radius( const double radius )
{
    if ( radius < _min_radius ) {
        _radius = _min_radius;
    }
    else if ( radius > _max_radius ) {
        _radius = _max_radius;
    }
    else {
        _radius = radius;
    }

    _update();
}

void Camera::set_elevation( const double elevation )
{
    if ( elevation < _min_elevation ) {
        _elevation = _min_elevation;
    }
    else if ( elevation > _max_elevation ) {
        _elevation = _max_elevation;
    }
    else {
        _elevation = elevation;
    }

    _update();
}

void Camera::set_azimuth( const double azimuth )
{
    _azimuth = fmod( azimuth, 360.0 );

    _update();
}

void Camera::apply_projection()
{
    GLuint w = CL_Display::get_width();
    GLuint h = CL_Display::get_height();

    glViewport( 0, 0, w, h );

    glMatrixMode( GL_PROJECTION );
    switch ( _project_type ) {
    case ORTHOGRAPHIC:
        gluOrtho2D( 0.0, w, 0.0, h );
        break;
    case PERSPECTIVE:
        // FIXME: parameterize, and pick better defaults
        gluPerspective( 45.0f, GLfloat( w ) / GLfloat( h ), 1.0f, 1000.0f );
        break;
    default:
        glLoadIdentity();
        break;
    }
}

void Camera::apply_view()
{
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glTranslated(0.0, 0.0, -_radius);
    glRotated(_elevation, 1.0, 0.0, 0.0);
    glRotated(-_azimuth, 0.0, 1.0, 0.0);

    // modelview = T*R*R.  first rotate around y axis, then about x axis,
    // then move camera straight backwards.  voila, spherical coordinates.
    // up: (0, 1, 0)
    // forward: (0, 0, 0) - position
}

void Camera::_update()
{
    // calculate position;
    // y = r*sin(_elevation)
    // x = r*cos(_elevation)*sin(_azimuth)
    // z = r*cos(_elevation)*cos(_azimuth)
    double rad_e = DEG_TO_RAD( _elevation );
    double rad_a = DEG_TO_RAD( _azimuth );
    double r_cos_e = _radius * cos( rad_e );

    _position = Vec3( r_cos_e * sin( rad_a ),
            _radius * sin( rad_e ),
            r_cos_e * cos( rad_a ) );

    // forward vec: 0 - pos
    _forward_vec = -get_position();
    unitize( _forward_vec );

    // up vec: always (0 1 0)

    // right vec:
    _right_vec = _forward_vec ^ _up_vec;
    unitize( _right_vec );
}

std::ostream& operator<<(std::ostream & os, const Camera & camera)
{
//    os << "radius: " << camera._radius
//        << ", angle (" << camera._azimuth
//        << ", " << camera._elevation << endl;
//    os << "position: " << camera.get_position() << endl;

    return os;
}
