#ifndef _CAMERA_H
#define _CAMERA_H

#include "main.h"

#include <ClanLib/display.h>

#include <gfx/vec3.h>

/**
 * A camera that always looks at the origin.
 *
 * Note: not quite spherical coords, because phi is usually the angle from
 * the vertical axis, whereas my equivalent "elevation" is the angle from
 * the horizontal plane (basically, 90 - phi).  Also, all the axes are
 * different (for ex, y is up here, not z).
 */
class Camera {
public:

    /// different types of projections the camera supports
    enum ProjectType {
        ORTHOGRAPHIC,
        PERSPECTIVE
    };

    Camera(ProjectType pt = PERSPECTIVE);

    void set_radius( const double radius );
    void set_elevation( const double elevation );
    void set_azimuth( const double azimuth );

    double get_radius( ) const { return _radius; }
    double get_elevation( ) const { return _elevation; }
    double get_azimuth( ) const { return _azimuth; }

    Vec3 get_position() const { return _position; }

    Vec3 get_right_vec() const { return _right_vec; }
    Vec3 get_up_vec() const { return _up_vec; }
    Vec3 get_forward_vec() const { return _forward_vec; }

    /**
     * Apply the camera's viewport and projection transformations.
     */
    void apply_projection();

    /**
     * Apply the camera's modelview transformations.
     */
    void apply_view();

    friend std::ostream& operator<<(std::ostream & os, const Camera& camera);
protected:

    // update the camera's internal cache.
    void _update();

    double _radius;          ///< distance of camera from origin
    double _elevation;       ///< vertical angle (phi in spherical coords)
    double _azimuth;         ///< horizontal angle (theta)

    ProjectType _project_type;   ///< orthographic or perspective

    double _min_radius;
    double _max_radius;
    double _min_elevation;
    double _max_elevation;

    Vec3 _position;
    Vec3 _right_vec;
    Vec3 _up_vec;
    Vec3 _forward_vec;
};

#endif // header guard
