/**
 * @file scene/interface.cpp
 * Implementation of Scene/interface class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "object.h"
#include "camera.h"
#include "utils.h"
#include "odeutils.h"
#include "gamestate.h"
#include "scene.h"

#include <ClanLib/display.h>
#include <gfx/arcball.h>

typedef TVec2<int> Vec2i;

bool out_of_bounds( Vec3 newpos )
{
    Vec3 world_size = GameState::get_gamestate()->get_level()->get_world_size() / 2.0;

    return (fabs( newpos[0] ) > world_size[0] ||
            fabs( newpos[1] ) > world_size[1] ||
            fabs( newpos[2] ) > world_size[2] );
}

void put_in_bounds( Vec3& newpos )
{
    Vec3 world_size = GameState::get_gamestate()->get_level()->get_world_size() / 2.0;

    for ( int i = 0; i < 3; i++ ) {
        if ( newpos[i] > world_size[i] )  newpos[i] = world_size[i];
        else if ( newpos[i] < -world_size[i] )    newpos[i] = -world_size[i];
    }
}

void mouse_drag_object( Camera *camera, Object *obj, Vec2 prev, Vec2 cur )
{
    bool editor_mode = SceneManager::get_scene()->is_editor_mode();

    if ( CL_Keyboard::get_keycode( CL_KEY_CONTROL ) ) {
        // rotation
        Quat quat;
        dQuaternion_to_quat( quat, dBodyGetQuaternion( obj->get_body_id() ) );

        double rot_angle = 2.0 * acos( quat.scalar() );
        Vec3 rot_axis = quat.vector();

        Vec3 up = camera->get_up_vec();
        Vec3 right = camera->get_right_vec();
        Vec3 forward = up ^ right;

        // invert the object's transform, so we rotate about the camera's
        // up vector.
        Mat4 inv_rot = rotation_matrix_rad( -rot_angle, rot_axis );

        Vec2 drag = 5.0 * Vec2( ( cur[ 0 ] - prev[ 0 ] ) / CL_Display::get_width(), 
                          ( cur[ 1 ] - prev[ 1 ] ) / CL_Display::get_height() );

        if ( CL_Mouse::get_keycode( CL_MOUSE_RIGHT ) ) {
            Quat qy = axis_to_quat( inv_rot * up, drag[ 0 ] );

            for ( int i = 0; i < 3; i++ ) {
                if ( !obj->is_rotatable( i ) && !editor_mode ) {
                    qy.vector()[i] = 0;
                }
            }

            quat = quat * qy;
        }
        else {
            Quat qx = axis_to_quat( inv_rot * right, drag[ 1 ] );
            Quat qz = axis_to_quat( inv_rot * forward, drag[ 0 ] );

            for ( int i = 0; i < 3; i++ ) {
                if ( !obj->is_rotatable( i ) && !editor_mode ) {
                    qx.vector()[i] = 0;
                    qz.vector()[i] = 0;
                }
            }

            quat = quat * qz * qx;
        }
        unitize( quat );

        dQuaternion dq;
        quat_to_dQuaternion( dq, quat );

        obj->set_orientation_quat( dq );
    }
    else {
        // translation
        if( !obj->is_movable() && !editor_mode )
            return;

        // goal: try to keep object as close to the mouse as possible.
        // this is achieved by retrieving the mouse's position in the
        // world, then moving the object to that position (after
        // constraining movement to be in the appropriate axes).  due to
        // movement constraint, there will be some error, so we do it a
        // bunch of times the object hasn't moved much.
        for (int i = 0; i < 10; i++) {
            Vec3 pos = obj->get_position();
            Vec3 mouse = world_to_screen( pos );
            mouse[0] = cur[0];
            mouse[1] = cur[1];

            Vec3 newpos = screen_to_world( mouse );

            if ( CL_Mouse::get_keycode( CL_MOUSE_RIGHT ) ) {
                // drag along y axis (ie, don't change x or z)
                newpos[0] = pos[0];
                newpos[2] = pos[2];
            }
            else {
                // drag in xz plane (ie, don't change y)
                newpos[1] = pos[1];
            }

            if ( out_of_bounds( newpos ) ) {
                put_in_bounds( newpos );
            }

            obj->set_position( newpos );

            if ( norm2( newpos - pos ) < 0.1 ) {
                // object didn't move much: we're close enough.
                break;
            }
        }
    }
}

void key_down_object(Camera *camera, Object* obj)
{
    const double amount = 0.1;

    bool editor_mode = SceneManager::get_scene()->is_editor_mode();
    Vec3 pos = obj->get_position();

    if ( obj->is_movable() || editor_mode ) {
        if ( CL_Keyboard::get_keycode( CL_KEY_S ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, amount, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_X ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, -amount, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_Z ) ) {
            obj->set_position( obj->get_position() + Vec3( -amount, 0.0, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_C ) ) {
            obj->set_position( obj->get_position() + Vec3( amount, 0.0, 0.0 ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_F ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, 0.0, -amount ) );
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_V ) ) {
            obj->set_position( obj->get_position() + Vec3( 0.0, 0.0, amount ) );
        }

        Vec3 pos = obj->get_position();
        if ( out_of_bounds( pos ) ) {
            put_in_bounds( pos );
            obj->set_position( pos );
        }
    }

    // rotation
    Quat quat;
    dQuaternion_to_quat( quat, dBodyGetQuaternion( obj->get_body_id() ) );

    if ( obj->is_rotatable( 0 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( CL_KEY_K ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( CL_KEY_M ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 1.0, 0.0, 0.0 ), x );
        }
    }

    if ( obj->is_rotatable( 1 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( CL_KEY_L ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( '.' ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 0.0, 1.0, 0.0 ), x );
        }
    }

    if ( obj->is_rotatable( 2 ) || editor_mode ) {
        double x = 0.0;
        if ( CL_Keyboard::get_keycode( ',' ) ) {
            x = amount;
        }
        if ( CL_Keyboard::get_keycode( '/' ) ) {
            x = -amount;
        }
        if ( x != 0.0 ) {
            quat = quat * axis_to_quat( Vec3( 0.0, 0.0, 1.0 ), x );
        }
    }

    dQuaternion dq;
    quat_to_dQuaternion( dq, quat );
    obj->set_orientation_quat( dq );
}

void key_down_camera(Camera *camera)
{
    const double dangle = 5.0;
    const double dradius = 1.0;

    if ( CL_Keyboard::get_keycode( CL_KEY_LEFT ) ) {
        camera->set_azimuth( camera->get_azimuth() - dangle );
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_RIGHT ) ) {
        camera->set_azimuth( camera->get_azimuth() + dangle );
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_DOWN ) ) {
        camera->set_elevation( camera->get_elevation() - dangle );
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_UP ) ) {
        camera->set_elevation( camera->get_elevation() + dangle );
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_PRIOR ) ) {
        camera->set_radius( camera->get_radius() - dradius );
    }
    if ( CL_Keyboard::get_keycode( CL_KEY_NEXT ) ) {
        camera->set_radius( camera->get_radius() + dradius );
    }
}

#if 0
void mouse_drag_camera(Camera *camera, Vec2 prev, Vec2 cur)
{
    Vec2 drag = Vec2( ( cur[ 0 ] - prev[ 0 ] ) / CL_Display::get_width(), 
                      ( cur[ 1 ] - prev[ 1 ] ) / CL_Display::get_height() );

    drag *= camera->get_radius();

    if ( CL_Mouse::get_keycode( CL_MOUSE_LEFT ) ) {
        // rotate about x/y
        camera->_azimuth += -drag[0];
        camera->_elevation += -drag[1];

        if ( camera->_elevation > 90 ) {
            camera->_elevation = 90;
        }
        else if ( camera->_elevation < 0 ) {
            camera->_elevation = 0;
        }
    }
    else if ( CL_Mouse::get_keycode( CL_MOUSE_RIGHT ) ) {
        // zoom in/out
        camera->_radius += drag[1];
    }
}
#endif
