/**
 * @file model.cpp
 * Implementation of Model class.
 *
 * Copyright 2003 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "model.h"

#include <gfx/gl.h>

Model::Model() :
    _material( grey )
{
    _material.reload_texture();
}

Model::Model(const Material& mat) :
    _material( mat )
{
    _material.reload_texture();
}

void Model::render()
{
    _material.make_current();
}

// Draw a wireframe axis-aligned box whose opposite corners are given by the
// points 'min' and 'max'.
void draw_wire_cube( const Vec3& min, const Vec3& max, const Vec3f& color )
{
    glPushAttrib( GL_LIGHTING_BIT | GL_ENABLE_BIT );
    glDisable( GL_LIGHTING );
    glDisable( GL_TEXTURE_2D );

    glColor3fv(color);

    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0],min[1],min[2]); glVertex3d(min[0],max[1],min[2]);
    glVertex3d(max[0],max[1],min[2]); glVertex3d(max[0],min[1],min[2]);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0],min[1],max[2]); glVertex3d(min[0],max[1],max[2]);
    glVertex3d(max[0],max[1],max[2]); glVertex3d(max[0],min[1],max[2]);
    glEnd();

    glBegin(GL_LINES);
    glVertex3d(min[0],min[1],min[2]); glVertex3d(min[0],min[1],max[2]);
    glVertex3d(min[0],max[1],min[2]); glVertex3d(min[0],max[1],max[2]);
    glVertex3d(max[0],max[1],min[2]); glVertex3d(max[0],max[1],max[2]);
    glVertex3d(max[0],min[1],min[2]); glVertex3d(max[0],min[1],max[2]);
    glEnd();

    glPopAttrib();
}

void draw_solid_cube( const Vec3& min, const Vec3& max, const bool draw_faces[6] )
{
    glBegin( GL_QUADS );

    if ( !draw_faces || draw_faces[ CUBE_FRONT ] ) {
        glNormal3d( 0, 0, -1 );
        glTexCoord2f( 0, 1 );
        glVertex3d( min[0], min[1], min[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( min[0], max[1], min[2] );
        glTexCoord2f( 1, 0 );
        glVertex3d( max[0], max[1], min[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( max[0], min[1], min[2] );
    }

    if ( !draw_faces || draw_faces[ CUBE_BACK ] ) {
        glNormal3d( 0, 0, 1 );
        glTexCoord2f( 0, 1 );
        glVertex3d( min[0], min[1], max[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( min[0], max[1], max[2] );
        glTexCoord2f( 1, 0 );
        glVertex3d( max[0], max[1], max[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( max[0], min[1], max[2] );
    }

    if ( !draw_faces || draw_faces[ CUBE_LEFT ] ) {
        glNormal3d( -1, 0, 0 );
        glTexCoord2f( 0, 1 );
        glVertex3d( min[0], min[1], min[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( min[0], max[1], min[2] );
        glTexCoord2f( 1, 0 );
        glVertex3d( min[0], max[1], max[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( min[0], min[1], max[2] );
    }

    if ( !draw_faces || draw_faces[ CUBE_RIGHT ] ) {
        glNormal3d( 1, 0, 0 );
        glTexCoord2f( 0, 1 );
        glVertex3d( max[0], min[1], min[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( max[0], max[1], min[2] );
        glTexCoord2f( 1, 0 );
        glVertex3d( max[0], max[1], max[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( max[0], min[1], max[2] );
    }

    if ( !draw_faces || draw_faces[ CUBE_TOP ] ) {
        glNormal3d( 0, 1, 0 );
        glTexCoord2f( 0, 1 );
        glVertex3d( min[0], max[1], min[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( min[0], max[1], max[2] );
        glTexCoord2f( 1, 0 );
        glVertex3d( max[0], max[1], max[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( max[0], max[1], min[2] );
    }

    if ( !draw_faces || draw_faces[ CUBE_BOTTOM ] ) {
        glNormal3d( 0, -1, 0 );
        glTexCoord2f( 1, 0 );
        glVertex3d( min[0], min[1], min[2] );
        glTexCoord2f( 1, 1 );
        glVertex3d( min[0], min[1], max[2] );
        glTexCoord2f( 0, 1 );
        glVertex3d( max[0], min[1], max[2] );
        glTexCoord2f( 0, 0 );
        glVertex3d( max[0], min[1], min[2] );
    }

    glEnd();
}

// TODO: texture coords
void draw_capped_cylinder( double l, double r, int quality )
{
    int i,j;
    double tmp,nx,ny,nz,start_nx,start_ny,a,ca,sa;
    // number of sides to the cylinder (divisible by 4):
    const int n = quality*4;

    l *= 0.5;
    a = double(M_PI*2.0)/double(n);
    sa = (double) sin(a);
    ca = (double) cos(a);

    // draw cylinder body
    ny=1; nz=0;		  // normal vector = (0,ny,nz)
    glBegin (GL_TRIANGLE_STRIP);
    for (i=0; i<=n; i++) {
        glNormal3d (ny,nz,0);
        glVertex3d (ny*r,nz*r,l);
        glNormal3d (ny,nz,0);
        glVertex3d (ny*r,nz*r,-l);
        // rotate ny,nz
        tmp = ca*ny - sa*nz;
        nz = sa*ny + ca*nz;
        ny = tmp;
    }
    glEnd();

    // draw first cylinder cap
    start_nx = 0;
    start_ny = 1;
    for (j=0; j<(n/4); j++) {
        // get start_n2 = rotated start_n
        double start_nx2 =  ca*start_nx + sa*start_ny;
        double start_ny2 = -sa*start_nx + ca*start_ny;
        // get n=start_n and n2=start_n2
        nx = start_nx; ny = start_ny; nz = 0;
        double nx2 = start_nx2, ny2 = start_ny2, nz2 = 0;
        glBegin (GL_TRIANGLE_STRIP);
        for (i=0; i<=n; i++) {
            glNormal3d (ny2,nz2,nx2);
            glVertex3d (ny2*r,nz2*r,l+nx2*r);
            glNormal3d (ny,nz,nx);
            glVertex3d (ny*r,nz*r,l+nx*r);
            // rotate n,n2
            tmp = ca*ny - sa*nz;
            nz = sa*ny + ca*nz;
            ny = tmp;
            tmp = ca*ny2- sa*nz2;
            nz2 = sa*ny2 + ca*nz2;
            ny2 = tmp;
        }
        glEnd();
        start_nx = start_nx2;
        start_ny = start_ny2;
    }

    // draw second cylinder cap
    start_nx = 0;
    start_ny = 1;
    for (j=0; j<(n/4); j++) {
        // get start_n2 = rotated start_n
        double start_nx2 = ca*start_nx - sa*start_ny;
        double start_ny2 = sa*start_nx + ca*start_ny;
        // get n=start_n and n2=start_n2
        nx = start_nx; ny = start_ny; nz = 0;
        double nx2 = start_nx2, ny2 = start_ny2, nz2 = 0;
        glBegin (GL_TRIANGLE_STRIP);
        for (i=0; i<=n; i++) {
            glNormal3d (ny,nz,nx);
            glVertex3d (ny*r,nz*r,-l+nx*r);
            glNormal3d (ny2,nz2,nx2);
            glVertex3d (ny2*r,nz2*r,-l+nx2*r);
            // rotate n,n2
            tmp = ca*ny - sa*nz;
            nz = sa*ny + ca*nz;
            ny = tmp;
            tmp = ca*ny2- sa*nz2;
            nz2 = sa*ny2 + ca*nz2;
            ny2 = tmp;
        }
        glEnd();
        start_nx = start_nx2;
        start_ny = start_ny2;
    }
}
