/**
 * @file scene/scene.cpp
 * Implementation of Scene/scene class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#include "scene.h"
#include "utils.h"
#include "effects.h"
#include "material.h"
#include "interface.h"

#include <gfx/gl.h>
#include <gfx/gltools.h>

#include <ClanLib/sound.h>

static dWorldID _world_id;
static dSpaceID _space_id;
static dJointGroupID _contact_group;

dWorldID get_world_id( void ) { return _world_id; }
dSpaceID get_space_id( void ) { return _space_id; }

SceneManager* SceneManager::_scene = NULL;

SceneManager::SceneManager( CL_GraphicContext* gc, GameState* gamestate ) :
    _selected_obj( NULL )
{
    _scene = this;

    _glstate = new CL_OpenGLState( gc );

    _gamestate = gamestate;

    gl_init();

    _current_level_id = -1;

    _axes_type = DRAW_ON_SELECTED;

    _world_id = dWorldCreate();

    // these should be set by the current level
    dWorldSetGravity( _world_id, 0.0f, -9.81f, 0.0f );
    dWorldSetERP( _world_id, 0.4f );
    dWorldSetCFM( _world_id, 1e-5f );

    _space_id = dSimpleSpaceCreate( 0 );

    _contact_group = dJointGroupCreate( 0 );

    _editor_mode = false;
    _show_movables = true;
}

SceneManager::~SceneManager( void )
{
    dWorldDestroy( _world_id );
    dSpaceDestroy( _space_id );

    dJointGroupEmpty( _contact_group );
    dJointGroupDestroy( _contact_group );

    delete _glstate;
}

int SceneManager::register_signals( void )
{
    // this is hacked: see GUI::mouse_down.
    // _slot_click = CL_Mouse::sig_key_down().connect( this, &SceneManager::on_mouse_down );

    _slot_move = CL_Mouse::sig_move().connect( this, &SceneManager::on_mouse_move );
    _slot_keydown = CL_Keyboard::sig_key_down().connect( this, &SceneManager::on_key_down );
    _slot_keyup = CL_Keyboard::sig_key_up().connect( this, &SceneManager::on_key_up );

    _ignore_move_signal = 0;

    return 0;
}

void SceneManager::gl_init( void )
{
    _glstate->set_active();

    glClearColor( 0.078f, 0.361f, 0.753f, 1.0f );
    // glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    glClearDepth( 1.0 );

    glDepthFunc( GL_LESS );
    glEnable( GL_DEPTH_TEST );

    glShadeModel( GL_SMOOTH );
    glEnable( GL_NORMALIZE );

    glEnable( GL_LINE_SMOOTH );

    //const Vec4f ambient_light( 0.3f, 0.3f, 0.3f, 1.0f );
    // Enable lighting and set up the lighting environment.  We specify a
    // global ambient glow and create two point lights.
    glEnable( GL_LIGHTING );
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, Vec4f( 0.3f, 0.3f, 0.3f, 1.0f ) );

    const Vec4f light0_pos( 0.0f, 10.0f, 0.0f, 0.0f );
    const Vec4f light0_rgb( 0.5f, 0.5f, 0.5f, 1.0f );

    glLightfv( GL_LIGHT0, GL_POSITION, light0_pos );
    glLightfv( GL_LIGHT0, GL_DIFFUSE, light0_rgb );
    glLightfv( GL_LIGHT0, GL_SPECULAR, light0_rgb );
    glEnable( GL_LIGHT0 );

    const Vec4f light1_pos( 0.0f, 5.0f, 10.0f, 0.0f );
    const Vec4f light1_rgb( 0.5f, 0.5f, 0.5f, 1.0f );

    glLightfv( GL_LIGHT1, GL_POSITION, light1_pos );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, light1_rgb );
    glLightfv( GL_LIGHT1, GL_SPECULAR, light1_rgb );
    glEnable( GL_LIGHT1 );

    glEnable( GL_TEXTURE_2D );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    _camera.apply_projection();
}

void SceneManager::generate_handlers( void )
{
    RegisterEffect::iterator eit;

    Level::ObjectList objects = _gamestate->get_level()->get_objects();
    Level::ObjectIterator it;

    for ( eit = RegisterEffect::begin(); eit != RegisterEffect::end(); eit++ ) {
        (*eit)->clear_handlers();
    }

    for ( it = objects.begin(); it != objects.end(); it++ ) {
        for ( eit = RegisterEffect::begin(); eit != RegisterEffect::end(); eit++ ) {
            (*eit)->add_handler( (*it).second );
        }
    }
}

void SceneManager::begin_level( void )
{
    if( _gamestate->get_current_state() == LEVEL_LOADING ) {
        generate_handlers();

        // camera should be in the Level class
        _camera.set_radius( _gamestate->get_level()->_camera_radius );
        _camera.set_elevation( _gamestate->get_level()->_camera_elevation );
        _camera.set_azimuth( _gamestate->get_level()->_camera_azimuth );

        dJointGroupEmpty( _contact_group );

        _current_level_id = _gamestate->get_level()->get_level_id();

        _selected_obj = NULL;
        _selected_colliding = false;
    }
}

void SceneManager::finish_level( void )
{
    if( _gamestate->get_current_state() == END_LEVEL ) {
        _current_level_id = -1;
    }
}

void SceneManager::near_callback( void* data, dGeomID o1, dGeomID o2 )
{
    CollisionFuncArg* collision_func_arg = (CollisionFuncArg*) data;

    const unsigned long N = 6;
    dContact contacts[ N ];

    const unsigned long n = dCollide( o1, o2, N, &contacts[ 0 ].geom, sizeof( dContact ) );

    for( unsigned long i = 0; i < n; ++i ) {
        // CHANGE: objects are now part of body's data (for multi-geom
        // bodies).
        dBodyID body_id1 = dGeomGetBody( contacts[ i ].geom.g1 );
        dBodyID body_id2 = dGeomGetBody( contacts[ i ].geom.g2 );
        Object* object1 = (Object*) ( body_id1 ? dBodyGetData( body_id1 ) : NULL );
        Object* object2 = (Object*) ( body_id2 ? dBodyGetData( body_id2 ) : NULL );

//        cerr << "obj1 " << object1 << endl;
//        cerr << "obj2 " << object2 << endl;

        if( GameState::get_gamestate()->get_current_state() == RUN_SIM ) {
            contacts[ i ].surface.mode = dContactBounce;
            contacts[ i ].surface.mu = 250;
            contacts[ i ].surface.bounce = 0.5f;
            contacts[ i ].surface.bounce_vel = 0.05f;

            if ( object1 ) {
                object1->collide( &contacts[ i ] );

                if( object1->is_stationary() ) {
                    body_id1 = 0;
                }
            }

            if ( object2 ) {
                object2->collide( &contacts[ i ] );

                if( object2->is_stationary() ) {
                    body_id2 = 0;
                }
            }   
        }
        else if( GameState::get_gamestate()->get_current_state() == GAMEPLAY ) {
            contacts[ i ].surface.mode = dContactBounce;
            contacts[ i ].surface.mu = 250;

            if ( object1 && object1 == get_scene()->get_selected_object() ) {
                get_scene()->_selected_colliding = true;
            }

            if ( object2 && object2 == get_scene()->get_selected_object() ) {
                get_scene()->_selected_colliding = true;
            }
        }

        if( collision_func_arg->attach_joints &&
            ( body_id1 != 0 || body_id2 != 0 ) ) {
            dJointID c = dJointCreateContact( _world_id, _contact_group, &contacts[ i ] );
            dJointAttach( c, body_id1, body_id2 );
        }
    }
}

void SceneManager::tick( double t )
{
    begin_level();

    if( _gamestate->get_current_state() == RUN_SIM ) {
        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        Level::ActorList actors = _gamestate->get_level()->get_actors();
        Level::ActorIterator ait;

        // every actor and object should push any non-ode based (magnetism,
        // wind, fear, etc.) effects for this tick into ode
        for ( ait = actors.begin(); ait != actors.end(); ait++ ) {
            (*ait).second->tick( t );
        }
        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {
            (*oit).second->tick( t );
        }

        CollisionFuncArg collision_func_arg = { true };

        dSpaceCollide( _space_id, &collision_func_arg, &SceneManager::near_callback );

        dWorldStep( _world_id, dReal( t ) );

        dJointGroupEmpty( _contact_group );
    }
    else if( _gamestate->get_current_state() == GAMEPLAY ) {
        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        // TODO: this should probably only be called on GAMEPLAY->RUN_SIM
        // transition. unset the collision flags some other way.
        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {
            (*oit).second->start_sim();
        }
    }

    finish_level();
}

void SceneManager::render()
{
    _glstate->set_active();
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    if( _gamestate->get_current_state() == RUN_SIM  || 
        _gamestate->get_current_state() == GAMEPLAY ||
        _gamestate->get_current_state() == LEVEL_LOADING )
    {
        Level::ObjectList objects = _gamestate->get_level()->get_objects();
        Level::ObjectIterator oit;

        _camera.apply_view();

        // render the worldbox, should the obj->render calls be moved inside
        // this function as well?
        // answer: yes. move this whole function into level
        _gamestate->get_level()->render();

        for ( oit = objects.begin(); oit != objects.end(); oit++ ) {

            Object *obj = (*oit).second;
            glLoadName( GLuint( obj ) );
            obj->render();

            if ( ( obj == get_selected_object() ) ) {
                if ( _selected_colliding ) {
                    obj->draw_decor( Vec3( 1.0, 1.0, 0.0 ) ); // yellow
                }
                if ( obj->is_movable() ) {
                    obj->draw_decor( Vec3( 0.0, 1.0, 0.0 ) ); // green
                }
                else {
                    obj->draw_decor( Vec3( 1.0, 0.0, 0.0 ) ); // red
                }
            }
            else if ( _show_movables && obj->is_movable() ) {
                obj->draw_decor( Vec3( 0.0, 0.0, 1.0 ) ); // blue
            }

            if ( _gamestate->get_current_state() == GAMEPLAY ) {
                draw_axes( obj );
            }
        }
    }
}

void SceneManager::draw_axes( Object* obj )
{
    double x_r = 0.02, y_r = 0.02, z_r = 0.02;
    double length = 2.0;

    switch ( _axes_type ) {
    case DRAW_ON_NONE:
        break;
    case DRAW_ON_SELECTED:
        if ( obj != _selected_obj ) {
            break;
        }
        // fall through
    case DRAW_ON_ALL:
        if ( obj == _selected_obj ) {
            length *= 2;
            // scale the axes to show which the object moves in
            if ( CL_Mouse::get_keycode( CL_MOUSE_RIGHT ) ) {
                y_r *= 2;
            }
            else {
                x_r *= 2;
                z_r *= 2;
            }
        }

        draw_cylinder( obj->get_position(), Vec3( 1, 0, 0 ), red, x_r, length );
        draw_cylinder( obj->get_position(), Vec3( 0, 1, 0 ), green, y_r, length );
        draw_cylinder( obj->get_position(), Vec3( 0, 0, 1 ), blue, z_r, length );

        break;
    default:
        break;
    }
}

void SceneManager::on_mouse_down( const CL_InputEvent& event )
{
    if( _gamestate->get_current_state() == GAMEPLAY ) {
        _glstate->set_active();

        if ( get_selected_object() ) {
            if ( event.id == CL_MOUSE_LEFT && !CL_Keyboard::get_keycode( CL_KEY_CONTROL ) ) {
                deselect_object();
            }
        }
        else if ( event.id == CL_MOUSE_LEFT || event.id == CL_MOUSE_RIGHT ) {
            int pos[] = { event.mouse_pos.x, event.mouse_pos.y };
            select_object( pos );
            _update_mouse_pos();
        }

        _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
        _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
    }

}

void SceneManager::on_mouse_move( const CL_InputEvent& event )
{
    if ( _ignore_move_signal > 0 ) {
        _ignore_move_signal--;
        _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
        _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
        return;
    }

    if( _gamestate->get_current_state() == GAMEPLAY ||
        _gamestate->get_current_state() == RUN_SIM ) {
        Vec2 cur_mouse_pos( event.mouse_pos.x, event.mouse_pos.y );
        Object *obj = get_selected_object();

        if ( obj && _gamestate->get_current_state() == GAMEPLAY ) {
            _glstate->set_active();
            mouse_drag_object( &_camera, obj, _prev_mouse_pos, cur_mouse_pos );

            _check_selected_colliding();
        }

        _prev_mouse_pos[ 0 ] = event.mouse_pos.x;
        _prev_mouse_pos[ 1 ] = event.mouse_pos.y;
    }
}

void SceneManager::on_key_down( const CL_InputEvent& event )
{
    if( _gamestate->get_current_state() == GAMEPLAY ||
        _gamestate->get_current_state() == RUN_SIM ) {
        key_down_camera( &_camera );

        Object *obj = get_selected_object();
        if ( obj && _gamestate->get_current_state() == GAMEPLAY ) {
            _glstate->set_active();
            key_down_object( &_camera, obj );

            _check_selected_colliding();
        }

        // TODO: this is a hack. level needs to update it's camera in order
        // to save it. move the camera into Level where it belongs, and
        // this will be fixed.
        _gamestate->get_level()->_camera_radius = _camera.get_radius();
        _gamestate->get_level()->_camera_elevation = _camera.get_elevation();
        _gamestate->get_level()->_camera_azimuth = _camera.get_azimuth();

        switch ( event.id ) {
        case CL_KEY_ESCAPE:
            deselect_object();
            break;

        case CL_KEY_1:
            _show_movables = !_show_movables;
            break;

        case CL_KEY_2:
            _axes_type = AxesType( (int(_axes_type) + 1) % int(DRAW_NUM_TYPES) );
            break;
        case CL_KEY_CONTROL:
            if ( _selected_obj ) {
                //CL_Mouse::show();
            }
            break;

        case CL_KEY_E:
            //_editor_mode = !_editor_mode;
            break;

        case CL_KEY_R:
            if( _selected_obj && ( is_editor_mode() ||
                        ( _selected_obj->is_rotatable( 0 ) &&
                          _selected_obj->is_rotatable( 1 ) &&
                          _selected_obj->is_rotatable( 2 ) ) ) ) {

                dQuaternion identity;

                dQSetIdentity( identity );

                // should be reading in the original orientation from the level file
                // actually I like this way better.
                _selected_obj->set_orientation_quat( identity );
            }

            break;
        default:
           _update_mouse_pos();
        }


//        if ( event.repeat_count == 0 ) {
//           _update_mouse_pos();
//        }
    }
}

void SceneManager::on_key_up( const CL_InputEvent& event )
{
    if( _gamestate->get_current_state() == GAMEPLAY || 
        _gamestate->get_current_state() == RUN_SIM ) {

        if ( event.id == CL_KEY_CONTROL ) {
            if ( _selected_obj ) {
				// CL_Mouse::hide();
                _update_mouse_pos();
            }

            if ( event.id == CL_KEY_ESCAPE ) {
                deselect_object();
            }
        }
    }
}

void SceneManager::select_object(int *pos)
{
    GLuint buffer[128]; // may need to adjust the size

    // let opengl know we want to pick an object
    begin_opengl_pick( pos, 1.0, buffer, 128 );

    // render the scene so opengl knows where things are
    _camera.apply_projection();
    render();

    // find out what object was picked
    GLuint select_idx = complete_opengl_pick( buffer );

    if ( select_idx != opengl_pick_nil ) {
        _selected_obj = (Object*) select_idx;
        // CL_Mouse::hide();
    }
}

void SceneManager::deselect_object( void )
{
    if ( !get_selected_object() ) {
        return;
    }

    if( get_selected_object()->is_movable() ) {
        _update_mouse_pos();
    }

    if ( _selected_colliding ) {
        CL_SoundBuffer sample( "resources/sound/error.wav" );
        sample.play();
        return;
    }

    // CL_Mouse::show();

    _selected_obj = NULL;
}

void SceneManager::_update_mouse_pos()
{
    if ( !_selected_obj ) {
        // no object? don't care where the mouse is
        return;
    }

    _glstate->set_active();

    Vec3 pos = _selected_obj->get_position();
    Vec3 mouse = world_to_screen( pos );

    // move the mouse, but make sure we don't treat it as a move event.
    // NOTE: clanlib crashes if you disconnect the signal inside the
    // callback.
    _ignore_move_signal = 5;
    CL_Mouse::set_position( int( mouse[0] ), int( mouse[1] ) );
}

void SceneManager::_check_selected_colliding()
{
    if ( !get_selected_object() ) {
        return;
    }

    CollisionFuncArg collision_func_arg = { false };

    // thread-safe? colliding/enabled status could be updated/used by 
    // Object::start_sim or SceneManager::tick
    _selected_colliding = false;
    dSpaceCollide( _space_id, &collision_func_arg, &SceneManager::near_callback );

    // don't let the user deselect an object if it is inside another object
    // why check if it's movable?
//    if( get_selected_object()->is_movable() &&
//        get_selected_object()->is_colliding() ) {
    if ( _selected_colliding ) {
        get_selected_object()->set_bbox_color( Vec3( 1.0, 1.0, 0.0 ) );
        return;
    }
    else {
        get_selected_object()->set_bbox_color( Vec3( 0.0, 1.0, 0.0 ) );
    }
}
