/**
 * @file utils.h
 * Header file for Utils class.
 *
 * Copyright 2003-2004 by GameBuilders.
 * Distributed WITHOUT WARRANTY under the GPL.  See COPYING for details.
 */
#ifndef _UTIL_H
#define _UTIL_H

#include "main.h"
#include "material.h"

#include <string>
using namespace std;

#include <gfx/mat4.h>
#include <gfx/vec3.h>

#include <ClanLib/core.h>

/// convert world coords to screen
Vec3 world_to_screen(Vec3 world);
/// convert screen coords to world
Vec3 screen_to_world(Vec3 screen);

/**
 * Make a rotation matrix out of a triplet of angles to rotate around each
 * axis.
 * Milkshape3D likes to store their rotations this way.
 */
Mat4 rotation_matrix_angles_rad( const Vec3& angles );

/**
 * Set the translation vectors for a transformation matrix.
 */
void set_translation( Mat4& mat, const Vec3& translate );

/**
 * Invert a translation*rotation matrix.
 * Optimized inverse for T*R matrix.
 */
Mat4 inverse_t_r( const Mat4& mat );

/**
 * Draw a cylinder at center along direction dir, with material mat, with
 * the given width and height.
 */
void draw_cylinder( Vec3 center, Vec3 dir, const Material& mat, double width, double height );


// some clanlib xml-related utils
inline string get_string( const CL_DomElement& element, string name )
{
    return element.get_attribute( name );
}

inline int get_int( const CL_DomElement& element, string name )
{
    return CL_String::to_int( element.get_attribute( name ) );
}

inline float get_float( const CL_DomElement& element, string name )
{
    return CL_String::to_float( element.get_attribute( name ) );
}

inline double get_double( const CL_DomElement& element, string name )
{
    return CL_String::to_double( element.get_attribute( name ) );
}

inline bool get_bool( const CL_DomElement& element, string name )
{
    return CL_String::to_bool( element.get_attribute( name ) );
}

inline Vec3 get_vec3( const CL_DomElement& element, string name )
{
    vector<string> tokens =
        CL_String::tokenize( element.get_attribute( name ), " " );

    return Vec3( CL_String::to_double( tokens[ 0 ] ), 
            CL_String::to_double( tokens[ 1 ] ), 
            CL_String::to_double( tokens[ 2 ] ) );
}

inline CL_DomElement get_element( CL_ResourceManager& manager, string resource )
{
    return manager.get_resource( resource ).get_element();
}

/// component-wise multiplication of Vec's
inline Vec3 vec_mult( Vec3 a, Vec3 b )
{
    return Vec3( a[0] * b[0], a[1] * b[1], a[2] * b[2] );
}

inline Vec3 vec_dropw( Vec4 a )
{
    return Vec3( a[0], a[1], a[2] );
}

inline string to_string( bool x )
{
    return CL_String::from_bool( x );
}

inline string to_string( int x )
{
    return CL_String::from_int( x );
}

inline string to_string( double x )
{
    return CL_String::from_double( x );
}

inline string to_string( Vec3 v )
{
    return CL_String::from_double(v[0]) + " " +
        CL_String::from_double(v[1]) + " " +
        CL_String::from_double(v[2]);
}

// convert a path to native system format
string convert_path( string path );

#endif // header guard
